/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jboss.wsf.stack.metro;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.xml.namespace.QName;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.ws.WebServiceException;
import javax.xml.ws.http.HTTPBinding;
import javax.xml.ws.soap.MTOMFeature;
import javax.xml.ws.soap.SOAPBinding;

import org.xml.sax.EntityResolver;

import com.sun.xml.ws.api.BindingID;
import com.sun.xml.ws.api.WSBinding;
import com.sun.xml.ws.api.server.Container;
import com.sun.xml.ws.api.server.InstanceResolver;
import com.sun.xml.ws.api.server.Invoker;
import com.sun.xml.ws.api.server.SDDocumentSource;
import com.sun.xml.ws.api.server.WSEndpoint;
import com.sun.xml.ws.api.streaming.XMLStreamReaderFactory;
import com.sun.xml.ws.binding.WebServiceFeatureList;
import com.sun.xml.ws.handler.HandlerChainsModel;
import com.sun.xml.ws.resources.ServerMessages;
import com.sun.xml.ws.resources.WsservletMessages;
import com.sun.xml.ws.server.EndpointFactory;
import com.sun.xml.ws.server.ServerRtException;
import com.sun.xml.ws.streaming.Attributes;
import com.sun.xml.ws.streaming.TidyXMLStreamReader;
import com.sun.xml.ws.streaming.XMLStreamReaderUtil;
import com.sun.xml.ws.transport.http.HttpAdapter;
import com.sun.xml.ws.transport.http.ResourceLoader;
import com.sun.xml.ws.transport.http.DeploymentDescriptorParser.AdapterFactory;
import com.sun.xml.ws.util.HandlerAnnotationInfo;
import com.sun.xml.ws.util.exception.LocatableWebServiceException;
import com.sun.xml.ws.util.xml.XmlUtil;

/**
 * A copy of DeploymentDescriptorParser
 * @see com.sun.xml.ws.transport.http.servlet.WSServletContextListener
 * @author WS Development Team
 * @author Kohsuke Kawaguchi
 * @author Thomas.Diesler@jboss.org
 * @since 10-May-2007
 */
public class DeploymentDescriptorParserExt<A>
{
   private static final Logger logger = Logger.getLogger(com.sun.xml.ws.util.Constants.LoggingDomain + ".server.http");

   public static final String NS_RUNTIME = "http://java.sun.com/xml/ns/jax-ws/ri/runtime";
   public static final String JAXWS_WSDL_DD_DIR = "WEB-INF/wsdl";
   public static final QName QNAME_ENDPOINTS = new QName(NS_RUNTIME, "endpoints");
   public static final QName QNAME_ENDPOINT = new QName(NS_RUNTIME, "endpoint");
   public static final String ATTR_VERSION = "version";
   public static final String ATTR_NAME = "name";
   public static final String ATTR_IMPLEMENTATION = "implementation";
   public static final String ATTR_WSDL = "wsdl";
   public static final String ATTR_SERVICE = "service";
   public static final String ATTR_PORT = "port";
   public static final String ATTR_URL_PATTERN = "url-pattern";
   public static final String ATTR_ENABLE_MTOM = "enable-mtom";
   public static final String ATTR_MTOM_THRESHOLD_VALUE = "mtom-threshold-value";
   public static final String ATTR_BINDING = "binding";
   public static final String ATTRVALUE_VERSION_1_0 = "2.0";

   private final Container container;
   private final ClassLoader classLoader;
   private final ResourceLoader loader;
   private final AdapterFactory<A> adapterFactory;

   private static final org.jboss.logging.Logger log = org.jboss.logging.Logger.getLogger(DeploymentDescriptorParserExt.class);

   /**
    * Endpoint names that are declared.
    * Used to catch double definitions.
    */
   private final Set<String> names = new HashSet<String>();

   /**
    * WSDL/schema documents collected from /WEB-INF/wsdl. Keyed by the system ID.
    */
   private final Map<String, SDDocumentSource> docs = new HashMap<String, SDDocumentSource>();

   /**
    *
    * @param cl
    *      Used to load service implementations.
    * @param loader
    *      Used to locate resources, in particular WSDL.
    * @param container
    *      Optional {@link Container} that {@link WSEndpoint}s receive.
    * @param adapterFactory
    *      Creates {@link HttpAdapter} (or its derived class.)
    */
   public DeploymentDescriptorParserExt(ClassLoader cl, ResourceLoader loader, Container container, AdapterFactory<A> adapterFactory) throws MalformedURLException
   {
      classLoader = cl;
      this.loader = loader;
      this.container = container;
      this.adapterFactory = adapterFactory;

      // toggle EJB and WEB deployments
      String resourcePath = "/META-INF/";
      try
      {
         URL resource = loader.getResource(resourcePath);
         InputStream inputStream = null;
         try
         {
            inputStream = resource.openStream();
         }
         finally
         {
            if(inputStream!=null)
               inputStream.close();
         }

         log.warn("bypass collectDocs(), it doesnt work for EJB endpoints");
      }
      catch (Exception e)
      {
         resourcePath = "/WEB-INF/";
         log.debug("Collecting docs from " + resourcePath);

         collectDocs(resourcePath+"wsdl/");
         logger.fine("metadata=" + docs);
      }

   }

   /**
    * Parses the {@code sun-jaxws.xml} file and configures
    * a set of {@link HttpAdapter}s.
    */
   public List<A> parse(String systemId, InputStream is)
   {
      XMLStreamReader reader = null;
      try
      {
         reader = new TidyXMLStreamReader(XMLStreamReaderFactory.create(systemId, is, true), is);
         XMLStreamReaderUtil.nextElementContent(reader);
         return parseAdapters(reader);
      }
      finally
      {
         if (reader != null)
         {
            try
            {
               reader.close();
            }
            catch (XMLStreamException e)
            {
               throw new ServerRtException("runtime.parser.xmlReader", e);
            }
         }
         try
         {
            is.close();
         }
         catch (IOException e)
         {
            // ignore
         }
      }
   }

   /**
    * Parses the {@code sun-jaxws.xml} file and configures
    * a set of {@link HttpAdapter}s.
    */
   public List<A> parse(File f) throws IOException
   {
      FileInputStream in = new FileInputStream(f);
      try
      {
         return parse(f.getPath(), in);
      }
      finally
      {
         in.close();
      }
   }

   /**
    * Get all the WSDL & schema documents recursively.
    */
   private void collectDocs(String dirPath) throws MalformedURLException
   {
      Set<String> paths = loader.getResourcePaths(dirPath);
      if (paths != null)
      {
         for (String path : paths)
         {
            if (path.endsWith("/"))
            {
               collectDocs(path);
            }
            else
            {
               URL res = loader.getResource(path);
               docs.put(res.toString(), SDDocumentSource.create(res));
            }
         }
      }
   }

   private List<A> parseAdapters(XMLStreamReader reader)
   {
      if (!reader.getName().equals(QNAME_ENDPOINTS))
      {
         failWithFullName("runtime.parser.invalidElement", reader);
      }

      List<A> adapters = new ArrayList<A>();

      Attributes attrs = XMLStreamReaderUtil.getAttributes(reader);
      String version = getMandatoryNonEmptyAttribute(reader, attrs, ATTR_VERSION);
      if (!version.equals(ATTRVALUE_VERSION_1_0))
      {
         failWithLocalName("runtime.parser.invalidVersionNumber", reader, version);
      }

      while (XMLStreamReaderUtil.nextElementContent(reader) != XMLStreamConstants.END_ELEMENT)
         if (reader.getName().equals(QNAME_ENDPOINT))
         {

            attrs = XMLStreamReaderUtil.getAttributes(reader);
            String name = getMandatoryNonEmptyAttribute(reader, attrs, ATTR_NAME);
            if (!names.add(name))
            {
               logger.warning(WsservletMessages.SERVLET_WARNING_DUPLICATE_ENDPOINT_NAME(/*name*/));
            }

            String implementationName = getMandatoryNonEmptyAttribute(reader, attrs, ATTR_IMPLEMENTATION);
            Class<?> implementorClass = getImplementorClass(implementationName, reader, this.classLoader);
            EndpointFactory.verifyImplementorClass(implementorClass);

            SDDocumentSource primaryWSDL = getPrimaryWSDL(reader, attrs, implementorClass, this.loader, this.docs);

            QName serviceName = getQNameAttribute(attrs, ATTR_SERVICE);
            if (serviceName == null)
               serviceName = EndpointFactory.getDefaultServiceName(implementorClass);

            QName portName = getQNameAttribute(attrs, ATTR_PORT);
            if (portName == null)
               portName = EndpointFactory.getDefaultPortName(serviceName, implementorClass);

            //get enable-mtom attribute value
            String enable_mtom = getAttribute(attrs, ATTR_ENABLE_MTOM);
            String mtomThreshold = getAttribute(attrs, ATTR_MTOM_THRESHOLD_VALUE);
            String bindingId = getAttribute(attrs, ATTR_BINDING);
            if (bindingId != null)
               // Convert short-form tokens to API's binding ids
               bindingId = getBindingIdForToken(bindingId);
            WSBinding binding = createBinding(bindingId, implementorClass, enable_mtom, mtomThreshold);
            String urlPattern = getMandatoryNonEmptyAttribute(reader, attrs, ATTR_URL_PATTERN);

            // TODO use 'docs' as the metadata. If wsdl is non-null it's the primary.

            boolean handlersSetInDD = setHandlersAndRoles(binding, reader, serviceName, portName, this.classLoader);

            ensureNoContent(reader);
            Invoker invoker = createInvoker(implementorClass);
            WSEndpoint<?> endpoint = WSEndpoint.create(implementorClass, !handlersSetInDD, invoker,
              serviceName, portName, container, binding, primaryWSDL, docs.values(), createEntityResolver(this.loader), false);
            adapters.add(adapterFactory.createAdapter(name, urlPattern, endpoint));
         }
         else
         {
            failWithLocalName("runtime.parser.invalidElement", reader);
         }
      return adapters;
   }

   protected Invoker createInvoker(Class<?> implementorClass)
   {
      Invoker invoker = InstanceResolver.createDefault(implementorClass).createInvoker();
      return invoker;
   }

   /**
    * @param ddBindingId
    *      binding id explicitlyspecified in the DeploymentDescriptor or parameter
    * @param implClass
    *      Endpoint Implementation class
    * @param mtomEnabled
    *      represents mtom-enabled attribute in DD
    * @param mtomThreshold
    *      threshold value specified in DD
    * @return
    *      is returned with only MTOMFeature set resolving the various precendece rules
    */
   private static WSBinding createBinding(String ddBindingId, Class<?> implClass, String mtomEnabled, String mtomThreshold)
   {
      // Features specified through DD
      WebServiceFeatureList features;

      MTOMFeature mtomfeature = null;
      if (mtomEnabled != null)
      {
         if (mtomThreshold != null)
            mtomfeature = new MTOMFeature(Boolean.valueOf(mtomEnabled), Integer.valueOf(mtomThreshold));
         else mtomfeature = new MTOMFeature(Boolean.valueOf(mtomEnabled));
      }

      BindingID bindingID;
      if (ddBindingId != null)
      {
         bindingID = BindingID.parse(ddBindingId);
         features = bindingID.createBuiltinFeatureList();

         if (checkMtomConflict(features.get(MTOMFeature.class), mtomfeature))
         {
            throw new ServerRtException(ServerMessages.DD_MTOM_CONFLICT(ddBindingId, mtomEnabled));
         }
      }
      else
      {
         bindingID = BindingID.parse(implClass);
         // Since bindingID is coming from implclass,
         // mtom through Feature annotation or DD takes precendece

         features = new WebServiceFeatureList();
         if (mtomfeature != null)
            features.add(mtomfeature); // this wins over MTOM setting in bindingID
         features.addAll(bindingID.createBuiltinFeatureList());
      }

      return bindingID.createBinding(features.toArray());
   }

   private static boolean checkMtomConflict(MTOMFeature lhs, MTOMFeature rhs)
   {
      if (lhs == null || rhs == null)
         return false;
      return lhs.isEnabled() ^ rhs.isEnabled();
   }

   /**
    * JSR-109 defines short-form tokens for standard binding Ids. These are
    * used only in DD. So stand alone deployment descirptor should also honor
    * these tokens. This method converts the tokens to API's standard
    * binding ids
    *
    * @param lexical binding attribute value from DD. Always not null
    *
    * @return returns corresponding API's binding ID or the same lexical
    */
   private static String getBindingIdForToken(String lexical)
   {
      if (lexical.equals("##SOAP11_HTTP"))
      {
         return SOAPBinding.SOAP11HTTP_BINDING;
      }
      else if (lexical.equals("##SOAP11_HTTP_MTOM"))
      {
         return SOAPBinding.SOAP11HTTP_MTOM_BINDING;
      }
      else if (lexical.equals("##SOAP12_HTTP"))
      {
         return SOAPBinding.SOAP12HTTP_BINDING;
      }
      else if (lexical.equals("##SOAP12_HTTP_MTOM"))
      {
         return SOAPBinding.SOAP12HTTP_MTOM_BINDING;
      }
      else if (lexical.equals("##XML_HTTP"))
      {
         return HTTPBinding.HTTP_BINDING;
      }
      return lexical;
   }

   /**
    * Checks the deployment descriptor or {@link @WebServiceProvider} annotation
    * to see if it points to any WSDL. If so, returns the {@link SDDocumentSource}.
    *
    * @return
    *      The pointed WSDL, if any. Otherwise null.
    */
   private static SDDocumentSource getPrimaryWSDL
   (
      XMLStreamReader xsr,
      Attributes attrs,
      Class<?> implementorClass,
      ResourceLoader loader,
      Map<String, SDDocumentSource> docs
   )
   {
      String wsdlFile = getAttribute(attrs, ATTR_WSDL);
      if (wsdlFile == null)
      {
         wsdlFile = EndpointFactory.getWsdlLocation(implementorClass);
      }

      if (wsdlFile != null)
      {
    	 URL wsdl = null;
    	 try
    	 {
            wsdl = loader.getResource(wsdlFile);
    	 }
    	 catch (MalformedURLException ignore) {}
    	 
         if (wsdl == null)
         {
            throw new LocatableWebServiceException(ServerMessages.RUNTIME_PARSER_WSDL_NOT_FOUND(wsdlFile), xsr);
         }

         // bypass the docs init
         docs.put(wsdl.toExternalForm(), SDDocumentSource.create(wsdl));

         SDDocumentSource docInfo = docs.get(wsdl.toExternalForm());
         assert docInfo != null;
         return docInfo;
      }

      return null;
   }

   /**
    * Creates an {@link EntityResolver} that consults {@code /WEB-INF/jax-ws-catalog.xml}.
    */
   private static EntityResolver createEntityResolver(ResourceLoader loader)
   {
      try
      {
         return XmlUtil.createEntityResolver(loader.getCatalogFile());
      }
      catch (MalformedURLException e)
      {
         throw new WebServiceException(e);
      }
   }

   private static String getAttribute(Attributes attrs, String name)
   {
      String value = attrs.getValue(name);
      if (value != null)
      {
         value = value.trim();
      }
      return value;
   }

   private static QName getQNameAttribute(Attributes attrs, String name)
   {
      String value = getAttribute(attrs, name);
      if (value == null || value.equals(""))
      {
         return null;
      }
      else
      {
         return QName.valueOf(value);
      }
   }

   /*
   private static String getNonEmptyAttribute(XMLStreamReader reader, Attributes attrs, String name)
   {
      String value = getAttribute(attrs, name);
      if (value != null && value.equals(""))
      {
         failWithLocalName("runtime.parser.invalidAttributeValue", reader, name);
      }
      return value;
   }

   private static String getMandatoryAttribute(XMLStreamReader reader, Attributes attrs, String name)
   {
      String value = getAttribute(attrs, name);
      if (value == null)
      {
         failWithLocalName("runtime.parser.missing.attribute", reader, name);
      }
      return value;
   }*/

   private static String getMandatoryNonEmptyAttribute(XMLStreamReader reader, Attributes attributes, String name)
   {
      String value = getAttribute(attributes, name);
      if (value == null)
      {
         failWithLocalName("runtime.parser.missing.attribute", reader, name);
      }
      else if (value.equals(""))
      {
         failWithLocalName("runtime.parser.invalidAttributeValue", reader, name);
      }
      return value;
   }

   /**
    * Parses the handler and role information and sets it
    * on the {@link WSBinding}.
    * @return true if <handler-chains> element present in DD false otherwise.
    */
   private static boolean setHandlersAndRoles(WSBinding binding, XMLStreamReader reader, QName serviceName, QName portName, ClassLoader classLoader)
   {
      if (XMLStreamReaderUtil.nextElementContent(reader) == XMLStreamConstants.END_ELEMENT || !reader.getName().equals(HandlerChainsModel.QNAME_HANDLER_CHAINS))
      {
         return false;
      }

      HandlerAnnotationInfo handlerInfo = HandlerChainsModel.parseHandlerFile(reader, classLoader, serviceName, portName, binding);
      binding.setHandlerChain(handlerInfo.getHandlers());
      if (binding instanceof SOAPBinding)
      {
         ((SOAPBinding)binding).setRoles(handlerInfo.getRoles());
      }

      // move past </handler-chains>
      XMLStreamReaderUtil.nextContent(reader);
      return true;
   }

   private static void ensureNoContent(XMLStreamReader reader)
   {
      if (reader.getEventType() != XMLStreamConstants.END_ELEMENT)
      {
         fail("runtime.parser.unexpectedContent", reader);
      }
   }

   private static void fail(String key, XMLStreamReader reader)
   {
      logger.log(Level.SEVERE, key + reader.getLocation().getLineNumber());
      throw new ServerRtException(key, Integer.toString(reader.getLocation().getLineNumber()));
   }

   private static void failWithFullName(String key, XMLStreamReader reader)
   {
      throw new ServerRtException(key, reader.getLocation().getLineNumber(), reader.getName());
   }

   private static void failWithLocalName(String key, XMLStreamReader reader)
   {
      throw new ServerRtException(key, reader.getLocation().getLineNumber(), reader.getLocalName());
   }

   private static void failWithLocalName(String key, XMLStreamReader reader, String arg)
   {
      throw new ServerRtException(key, reader.getLocation().getLineNumber(), reader.getLocalName(), arg);
   }

   /**
    * Loads the class of the given name.
    * @param reader used to report the source location information if there's any error
    */
   private static Class<?> getImplementorClass(String name, XMLStreamReader reader, ClassLoader classLoader)
   {
      try
      {
         return Class.forName(name, true, classLoader);
      }
      catch (ClassNotFoundException e)
      {
         logger.log(Level.SEVERE, e.getMessage(), e);
         throw new LocatableWebServiceException(ServerMessages.RUNTIME_PARSER_CLASS_NOT_FOUND(name), e, reader);
      }
   }
}
